import enum
import json
import logging

import requests
from orchestrator import inputstep
from orchestrator.config.assignee import Assignee
from orchestrator.domain.lifecycle import SubscriptionModel
from orchestrator.forms import FormPage, ReadOnlyField
from orchestrator.forms.validators import Accept, Label, LongText
from orchestrator.types import UUIDstr, State
from orchestrator.utils.json import json_dumps

from gso import settings
from gso.products.product_types.iptrunk import IptrunkProvisioning, Iptrunk

logger = logging.getLogger(__name__)


class CUDOperation(str, enum.Enum):
    """
    Enum for different C(R)UD operations that the provisioning proxy supports.
    Read is not applicable, hence these become CUD and not CRUD operations.
    """
    #: Creation is done with a POST request
    POST = 'POST'
    #: Updating is done with a PUT request
    PUT = 'PUT'
    #: Removal is done with a DELETE request
    DELETE = 'DELETE'


def _send_request(endpoint: str, parameters: dict, process_id: UUIDstr,
                  operation: CUDOperation):
    oss = settings.load_oss_params()
    pp_params = oss.PROVISIONING_PROXY
    assert pp_params

    callback_url = f'{settings.load_oss_params().GENERAL.public_hostname}' \
                   f'/api/processes/{process_id}/resume'
    logger.debug(f'[provisioning proxy] provisioning for process {process_id}')

    parameters.update({'callback': callback_url})
    url = f'{pp_params.scheme}://{pp_params.api_base}/api/{endpoint}'

    request = None

    if operation == CUDOperation.POST:
        request = requests.post(url, json=parameters)
    elif operation == CUDOperation.PUT:
        request = requests.put(url, json=parameters)
    elif operation == CUDOperation.DELETE:
        request = requests.delete(url, json=parameters)

    if request.status_code != 200:
        raise AssertionError(request.text)


def provision_device(
        subscription: SubscriptionModel,
        process_id: UUIDstr,
        dry_run: bool = True):
    parameters = {
        'dry_run': dry_run,
        'subscription': json.loads(json_dumps(subscription))
    }

    _send_request('device', parameters, process_id, CUDOperation.POST)


def provision_ip_trunk(subscription: IptrunkProvisioning,
                       process_id: UUIDstr,
                       dry_run: bool = True):
    parameters = {
        'subscription': json.loads(json_dumps(subscription)),
        'dry_run': dry_run,
        'verb': '',  # FIXME
        'object': ''  # FIXME
    }

    _send_request('ip_trunk', parameters, process_id, CUDOperation.POST)


def modify_ip_trunk(old_subscription: Iptrunk,
                    new_subscription: Iptrunk,
                    process_id: UUIDstr,
                    dry_run: bool = True):
    parameters = {
        'dry_run': dry_run,
        'old_subscription': old_subscription,
        'subscription': new_subscription
        # FIXME missing parameters
    }

    _send_request('ip_trunk', parameters, process_id, CUDOperation.PUT)


@inputstep('Await provisioning proxy results', assignee=Assignee('SYSTEM'))
def await_pp_results() -> State:
    class ProvisioningResultPage(FormPage):
        class Config:
            title = 'Do NOT click on confirm in this step!'

        warning_label: Label = 'This step relies on an external service to ' \
                               'send an update to the orchestrator, do not ' \
                               'interfere with this process please.'
        pp_run_results: dict = {'state': 'not_ready'}
        confirm: Accept = Accept('INCOMPLETE')

    result_page = yield ProvisioningResultPage

    return result_page.dict()


@inputstep('Confirm provisioning proxy results', assignee=Assignee('SYSTEM'))
def confirm_pp_results(state: State) -> State:
    class ConfirmRunPage(FormPage):
        class Config:
            title = 'Execution completed, please confirm the results.'

        run_status: str = ReadOnlyField(state['pp_run_results']['status'])
        run_results: LongText = ReadOnlyField(
            f"{state['pp_run_results']['output']}")
        confirm: Accept = Accept('INCOMPLETE')

    yield ConfirmRunPage

    return state
